@import url("https://fonts.googleapis.com/css?family=Roboto+Mono");

dd {
    margin-left: 30px;
}

.title {
	font-size: 200%;
	font-weight: bold;
}

.title .ingroups {
	font-size: 50%;
}

h1 {
	font-size: 150%;
	color: #354C7B;
	background: none;
	border-bottom: 1px solid #879ECB;
	font-size: 150%;
	font-weight: normal;
	padding-top: 8px;
	padding-bottom: 8px;
	padding-left: 0px;
	width: 100%;
}

h2 {
	font-size: 120%;
	color: #354C7B;
	background: none;
	border-bottom: 1px solid #879ECB;
	font-size: 150%;
	font-weight: normal;
	padding-top: 8px;
	padding-bottom: 8px;
	padding-left: 0px;
	width: 100%;
}

.sm-dox li {
	float:left;
	border-top: 0;
	padding-right: 20px;
}

.sm li, .sm a {
	position: relative;
}

.sm, .sm ul, .sm li {
	list-style: none;
	display: block;
	line-height: normal;
	direction: ltr;
	text-align: left;
}

.sm, .sm *, .sm *::before, .sm *::after {
	box-sizing: border-box;
}

#main-nav {
	padding: 30px;
}

/* Main menu sub-items like file-list, etc */
#main-menu li ul {
	display: none;
}

.paramname {
	padding-right: 10px;
}

.memtitle {
	background-image: none;
	background-color: #F0F0F0;
}

.memproto {
	background-color: #F0F0F0;
}

.headertitle {
	background-image: none;
	background-color: #F0F0F0;
}

div.header {
	border: none;
}

td.fieldname {
	font-family: 'Roboto Mono', monospace;
}

.fieldtable th {
	background-image: none;
	background-color: #F0F0F0;
}

body {
	letter-spacing: 0px;
}

.mdescLeft, .mdescRight, .memItemLeft, .memItemRight, .memTemplItemLeft, .memTemplItemRight, .memTemplParams {
	background-color: #F0F0F0;
}

a {
	color: #2873b0;

}

.navpath ul {
	background-image: none;
	background-color: #F0F0F0;
}
